// Copyright (c) Microsoft Corporation.
// Licensed under the MIT License.

#ifndef _SCENEPIC_H_
#define _SCENEPIC_H_

#include "io.h"
#include "loop_subdivision_stencil.h"
#include "scene.h"
#include "transforms.h"

#endif