// Copyright (c) Microsoft Corporation.
// Licensed under the MIT License.

#include "json_value.h"

#include "json/json.h"

namespace scenepic
{
  Json::Value scenepic_to_json(const JsonValue& value);
}